#include <cstdio>
#include <cstdlib>
#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  pragma comment(lib, "glew32.lib")
#  include "glew.h"
#  include "glut.h"
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

/*
** ο
*/
#define POINTS 100000

/*
**ȥåܡ
*/
#include "Trackball.h"
static Trackball trackball;
static int pressedButton;

/*
** Ѵ
*/
#include "Matrix.h"
static Matrix projectionMatrix;
static Matrix modelviewMatrix;
static GLint transformMatrixLocation;

/*
** 
*/
#include "shadersource.h"
static const char vertSource[] = "simple.vert";
static const char fragSource[] = "simple.frag";
static GLuint gl2Program;

/*
** ޷
*/
#include "disseminate.h"
static GLuint buffer[1];
static GLsizei points;
static GLint pointLocation;

/*
** ɽ
*/
static void display(void)
{
  /* ѴѴǥѴ */
  Matrix transformMatrix = projectionMatrix * modelviewMatrix * trackball.get();
  transformMatrix.scale(1.0f, 1.0f, 2.0f);
  transformMatrix.translate(0.0f, 0.0f, -0.5f);
  
  /* ץŬѤ */
  glUseProgram(gl2Program);

  /* uniform ѿ transformMatrix ˹ꤹ */
  glUniformMatrix4fv(transformMatrixLocation, 1, GL_FALSE, transformMatrix.get());

  /* attribute ѿ point ĺбդ */
  glEnableVertexAttribArray(pointLocation);

  /* ĺХåե֥ȤȤ buffer[0] ꤹ */
  glBindBuffer(GL_ARRAY_BUFFER, buffer[0]);

  /* ĺγǼȽ񼰤ꤹ */
  glVertexAttribPointer(pointLocation, 3, GL_FLOAT, GL_FALSE, 0, 0);

  /* ̥ꥢ */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* ̾õͭˤ */
  glEnable(GL_DEPTH_TEST);

  /* ޷ */
  glDrawArrays(GL_POINTS, 0, points);

  /* ̾õ̵ˤ */
  glDisable(GL_DEPTH_TEST);

  /* ĺХåե֥Ȥ */
  glBindBuffer(GL_ARRAY_BUFFER, 0);

  /* index  0  attribute ѿĺȤбդ */
  glDisableVertexAttribArray(pointLocation);

  glutSwapBuffers();
}

/*
** ɥ
*/
static void resize(int w, int h)
{
  /* ӥ塼ݡ */
  glViewport(0, 0, w, h);

  /* ƩѴ */
  projectionMatrix.loadIdentity();
  projectionMatrix.camera(30.0f, (GLfloat)w / (GLfloat)h, 5.0f, 9.0f);
  
  /* ȥåܡ뤹ϰ */
  trackball.region(w, h);
}

/*
** ܡ
*/
static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
    case '\033':
    case 'q':
    case 'Q':
      exit(0);
    default:
      break;
  }
}

/*
** ˥᡼
*/
static void idle(void)
{
  /* ̤ؤ */
  glutPostRedisplay();
}

/*
** ޥΥå
*/
static void mouse(int button, int state, int x, int y)
{
  pressedButton = button;
  
  switch (pressedButton) {
    case GLUT_LEFT_BUTTON:
      if (state == GLUT_DOWN) {
        /* ȥåܡ볫 */
        trackball.start(x, y);
      }
      else {
        /* ȥåܡ */
        trackball.stop(x, y);
      }
      break;
    default:
      break;
  }
}

/*
** ޥΥɥå
*/
static void motion(int x, int y)
{
  switch (pressedButton) {
    case GLUT_LEFT_BUTTON:
      /* ȥåܡư */
      trackball.motion(x, y);
      break;
    default:
      break;
  }
}

/*
** 
*/
static void init(void)
{
  /* ץΥѥ롿󥯷̤ѿ */
  GLint compiled, linked;
  
#if defined(WIN32)
  /* GLEW ν */
  GLenum err = glewInit();
  if (err != GLEW_OK) {
    fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
    exit(1);
  }
#endif
  
  /* ֥Ȥκ */
  GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
  GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
  
  /* Υץɤ߹ */
  if (readShaderSource(vertShader, vertSource)) exit(1);
  if (readShaderSource(fragShader, fragSource)) exit(1);
  
  /* СƥåΥץΥѥ */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(vertShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in vertex shader.\n");
    exit(1);
  }
  
  /* ե饰ȥΥץΥѥ */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(fragShader);
  if (compiled == GL_FALSE) {
    fprintf(stderr, "Compile error in fragment shader.\n");
    exit(1);
  }
  
  /* ץ४֥Ȥκ */
  gl2Program = glCreateProgram();
  
  /* ֥ȤΥץؤϿ */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);
  
  /* ֥Ȥκ */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);
  
  /* ץΥ */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &linked);
  printProgramInfoLog(gl2Program);
  if (linked == GL_FALSE) {
    fprintf(stderr, "Link error.\n");
    exit(1);
  }
  
  /* uniform ѿ transformMatrix ξ */
  transformMatrixLocation = glGetUniformLocation(gl2Program, "transformMatrix");
  
  /* attribute ѿ point ξ */
  pointLocation = glGetAttribLocation(gl2Program, "point");
  
  /* Ѵ */
  modelviewMatrix.loadIdentity();
  modelviewMatrix.lookat(0.0f, 0.0f, 7.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
  
  /* ĺХåե֥Ȥ򣱤ĺ */
  glGenBuffers(1, buffer);
  
  /* ޷Хåե֥ȤϿ */
  points = disseminate(POINTS, buffer);
  
  /* ȥåܡν */
  trackball.initialize();
  
  /* طʿ */
  glClearColor(0.0, 0.1, 0.3, 1.0);
}

/*
** ᥤץ
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutIdleFunc(idle);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();

  return 0;
}
